
import tempfile

from gensim.corpora import Dictionary



def convert_gs_dictionary_to_string(dictionary: Dictionary) -> str:
    """Convert corpora to string"""
    with tempfile.NamedTemporaryFile() as fp:
        dictionary.save_as_text(fp.file.name) # , sort_by_word=True) # save to text
        return fp.read().decode()

def convert_string_to_gs_dictionary(s: str) -> Dictionary:
    """Convert string to corpora"""
    with tempfile.NamedTemporaryFile() as fp:
        fp.write(s.encode('utf-8'))
        return Dictionary.load_from_text(fp.file.name)


""" Object format validation """

from typing import Optional, Any, Sequence, Callable, Union, Set
import datetime

import numpy as np


def is_datetime_formatted_str(s: Any, fmt: str) -> bool:
    """Check that string is date-formatted"""
    if not isinstance(s, str): # possibly redundant, but safer
        return False
    try:
        datetime.datetime.strptime(s, fmt)
        return True
    except Exception as e:
        return False

def is_list_of_strings(obj: Sequence) -> bool:
    """Check that object is a list of strings."""
    return isinstance(obj, list) and all([isinstance(e, str) for e in obj])

def is_list_of_floats(obj: Sequence) -> bool:
    """Check if object is a list of floats."""
    return isinstance(obj, list) and all([isinstance(val, (int, float)) for val in obj])

def is_list_of_list_of_strings(obj: Sequence) -> bool:
    """Check that iterable is a list of lists of strings."""
    return isinstance(obj, list) and all([is_list_of_strings(lst) for lst in obj])

def is_list_of_formatted_strings(obj: Sequence,
                                 fmt_check_func: Callable,
                                 list_len: Optional[int] = None) -> bool:
    """Check that iterable is list of formatted strings (formatting checked by provided function)."""
    if isinstance(obj, list):
        len_cond = True if list_len is None else len(obj) == list_len
        return len_cond and all([fmt_check_func(e) for e in obj])
    return False

def is_list_of_list_of_time_range_strings(obj: Sequence,
                                          func: Callable,
                                          num_ranges: Optional[int] = None) \
        -> bool:
    """Check that object is of the form [[<date_or_timestamp_string>, <date_or_timestamp_string>], ...]."""
    if isinstance(obj, list):
        len_cond = True if num_ranges is None else len(obj) == num_ranges
        return len_cond and all([is_list_of_formatted_strings(lst, func, list_len=2) for lst in obj])
    return False

def is_subset(obj1: Union[Sequence, Set],
              obj2: Union[Sequence, Set]) \
        -> bool:
    """
    Check that the elements in one iterable comprise a subset of the elements in the other.

    For example:
        assert_subset([0, 1, 2], [0, 4, 2, 1]) returns True
        assert_subset([0, 1, 2], [3, 2, 1, 5]) returns False
    """
    return len(set(obj1) - set(obj2)) == 0

def is_list_of_sequences(obj: Sequence,
                         seq_types: tuple,
                         len_: Optional[int] = None) \
        -> bool:
    """
    Check that object is a list of an iterable type (e.g. tuple, list), optionally checking the length of each entry.
    """
    if isinstance(obj, list):
        for e in obj:
            valid = isinstance(e, seq_types) and (True if len_ is None else len(e) == len_)
            if not valid:
                return False
        return True
    return False

def is_dict_of_instances(obj: Sequence,
                         type) \
        -> bool:
    """Check that object is a dict of objects of specified instance(s)."""
    return isinstance(obj, dict) and all([isinstance(val, type) for key, val in obj.items()])

def is_list_of_instances(obj: Sequence,
                         type) \
        -> bool:
    """Check that object is a list of objects of specified instance(s)."""
    return isinstance(obj, list) and all([isinstance(val, type) for val in obj])

def is_int_or_float(val) -> bool:
    return isinstance(val, (int, float, np.int64, np.float64))



# TODO: add tests for this module


from typing import Optional

import numpy as np
from scipy.interpolate import interp1d


def resample_arr(ts: np.ndarray,
                 data: np.ndarray,
                 period: float,
                 max_ts_gap: Optional[float] = None,
                 omit_zeroes: bool = False):
    """Resample columns of an array individually, handling issues with large gaps and zero values."""
    ts_resamp = np.arange(np.min(ts), np.max(ts), period) # starts at 0

    num_samps_resamp = len(ts_resamp)
    num_cols = data.shape[1]
    data_resamp = np.zeros((num_samps_resamp, num_cols), dtype='float')
    for i in range(num_cols):
        # remove zeroes (always keep important values)
        if omit_zeroes:
            mask: np.ndarray = (data[:, i] != 0)
            mask[0] = True
            ts_i = ts[mask]
            data_i = data[mask, i]
        else:
            ts_i = ts
            data_i = data[:, i]

        # interpolate
        if len(ts_i) <= 1:
            continue
        kind = 'cubic' if len(ts_i) >= 4 else 'linear'
        f = interp1d(ts_i, data_i, kind=kind, axis=0, fill_value="extrapolate") # , assume_sorted=True)#, copy=False)
        data_resamp[:, i] = f(ts_resamp)

        # replace spline interp over long gaps with linear interp
        if max_ts_gap is not None:
            ts_diff = ts_i[1:] - ts_i[:-1]
            idxs_gap = np.where(ts_diff >= max_ts_gap)[0] # left indices at large gaps
            for j in idxs_gap:
                j_resamp_left = np.argmin(np.abs(ts_resamp - ts_i[j])) # closest left and right indices in ts_resamp
                j_resamp_right = np.argmin(np.abs(ts_resamp - ts_i[j + 1]))
                num_samps = j_resamp_right - j_resamp_left + 1
                ii = np.linspace(1, 0, num_samps)
                data_resamp[j_resamp_left:j_resamp_right + 1, i] = ii * data_i[j] + (1 - ii) * data_i[j + 1]

    return ts_resamp, data_resamp
